/** @param {NS} ns */
export async function main(ns) {
	const hackMinMoneyLimitPercent = 0.75
	const hackMaxSecurityIncreaseLimit = 5
	const hackMinLevelThreshPercent = 1
	const maxTargets = Infinity
	const loopSettings = {
		"poolBlacklist": [],
		// "poolBlacklist": ["home"],

		"moneyTargetsWhitelist": [],
		"moneyTargetsWhitelist": ["n00dles", "harakiri-sushi", "phantasy"],
		// "moneyTargetsBlacklist": [],
		"moneyTargetsBlacklist": ["joesguns"], // blacklist joesguns to avoid conflict with exp farm

		"expTargetHostname": "joesguns",
		"expMaxAvailRamUsePercent": 0.95,
	}

	const hackRamUsage = ns.getScriptRam("/payloads/hack.js")
	const growRamUsage = ns.getScriptRam("/payloads/grow.js")
	const weakenRamUsage = ns.getScriptRam("/payloads/weaken.js")
	const autoRackRamUsage = ns.getScriptRam("/payloads/autorack.js")
	const autoRackPwnUsage = ns.getScriptRam("/payloads/autopwn.js")

	const moneyFormat = new Intl.NumberFormat({ style: "currency" })

	ns.clearLog()
	ns.tail()
	ns.disableLog('ALL')
	// ns.disableLog('getServerUsedRam')
	// ns.disableLog('getServerMaxRam')
	// ns.print(`${pwnMethods.length} pwning method(s) available!`)

	class PoolException {
		constructor(message) {
			this.message = message
		}

		toString() {
			return `WARNING: ${this.message}`
		}
	}

	class Server {
		constructor(hostname) {
			this.hostname = hostname
			ns.scp(["/payloads/grow.js", "/payloads/hack.js", "/payloads/weaken.js", "/payloads/autorack.js", "/payloads/autopwn.js"], this.hostname)
		}

		toString() {
			return this.hostname
		}

		get hasRootAccess() {
			return ns.hasRootAccess(this.hostname)
		}

		get hackable() {
			return this.hasRootAccess && this.maxMoney > 0 && (this.requiredHackLevel <= ns.getHackingLevel() * hackMinLevelThreshPercent || this.requiredHackLevel == 1)
		}

		get requiredHackLevel() {
			return ns.getServerRequiredHackingLevel(this.hostname)
		}

		get maxRam() {
			return ns.getServerMaxRam(this.hostname)
		}

		get usedRam() {
			return ns.getServerUsedRam(this.hostname)
		}

		get availRam() {
			return this.maxRam - this.usedRam
		}

		get maxMoney() {
			return ns.getServerMaxMoney(this.hostname)
		}

		get availMoney() {
			return ns.getServerMoneyAvailable(this.hostname)
		}

		get minSecurityLevel() {
			return ns.getServerMinSecurityLevel(this.hostname)
		}

		get currSecurityLevel() {
			return ns.getServerSecurityLevel(this.hostname)
		}

		get hackTime() {
			return Math.ceil(ns.getHackTime(this.hostname))
		}

		get growTime() {
			return Math.ceil(ns.getGrowTime(this.hostname))
		}

		get weakenTime() {
			return Math.ceil(ns.getWeakenTime(this.hostname))
		}

		get hackChance() {
			return ns.hackAnalyzeChance(this.hostname)
		}

		get pids() {
			return ns.ps(this.hostname).map(process => process.pid)
		}

		getScriptCapacity(script) {
			return Math.floor(this.availRam / ns.getScriptRam(script))
		}

		getHackCapacity() {
			return Math.floor(this.availRam / hackRamUsage)
		}

		getGrowCapacity() {
			return Math.floor(this.availRam / growRamUsage)
		}

		getWeakenCapacity() {
			return Math.floor(this.availRam / weakenRamUsage)
		}

		computeOptimalHackThreadCount() {
			const requiredThreads = Math.ceil(((this.availMoney - (this.maxMoney * hackMinMoneyLimitPercent)) / this.availMoney) / ns.hackAnalyze(this.hostname))
			let securityLimitedThreads = 0
			let predictedSecurityIncrease = 0
			while ((predictedSecurityIncrease + this.currSecurityLevel) < (this.minSecurityLevel + hackMaxSecurityIncreaseLimit)) {
				predictedSecurityIncrease = ns.hackAnalyzeSecurity(++securityLimitedThreads)
			}
			return Math.min(requiredThreads, securityLimitedThreads)
		}

		computeOptimalGrowThreadCount() {
			const requiredThreads = Math.ceil(ns.growthAnalyze(this.hostname, this.maxMoney / this.availMoney))
			// let securityLimitedThreads = 0
			// let predictedSecurityIncrease = 0
			// while ((predictedSecurityIncrease + this.currSecurityLevel) < (this.minSecurityLevel + hackMaxSecurityIncreaseLimit)) {
			// 	predictedSecurityIncrease = ns.growthAnalyzeSecurity(++securityLimitedThreads)
			// }
			// return Math.min(requiredThreads, securityLimitedThreads)
			return requiredThreads
		}

		computeOptimalWeakenThreadCount() {
			let requiredThreads = 0
			let prediction = this.currSecurityLevel
			while (prediction > this.minSecurityLevel) {
				prediction = this.currSecurityLevel - ns.weakenAnalyze(++requiredThreads)
			}
			return requiredThreads
		}

		execHack(target, numThreads = 1) {
			const pid = ns.exec("/payloads/hack.js", this.hostname, numThreads, crypto.randomUUID(), numThreads, target.hostname)
			if (pid == 0) {
				throw `ERROR: failed to run hack() from ${this.hostname}, target = ${target}, numThreads = ${numThreads}, capacity = ${this.getHackCapacity()}`
			}
			return new Task(this, pid, "hack", Date.now() + target.hackTime, target)
		}

		execGrow(target, numThreads = 1) {
			const pid = ns.exec("/payloads/grow.js", this.hostname, numThreads, crypto.randomUUID(), numThreads, target.hostname)
			if (pid == 0) {
				throw `ERROR: failed to run grow() from ${this.hostname}, target = ${target}, numThreads = ${numThreads}, capacity = ${this.getGrowCapacity()}`
			}
			return new Task(this, pid, "grow", Date.now() + target.growTime, target)
		}

		execWeaken(target, numThreads = 1) {
			const pid = ns.exec("/payloads/weaken.js", this.hostname, numThreads, crypto.randomUUID(), numThreads, target.hostname)
			if (pid == 0) {
				throw `ERROR: failed to run weaken() from ${this.hostname}, target = ${target}, numThreads = ${numThreads}, capacity = ${this.getHackCapacity()}`
			}
			return new Task(this, pid, "weaken", Date.now() + target.weakenTime, target)
		}

		execAutoRack() {
			const pid = ns.exec("/payloads/autorack.js", this.hostname, 1, crypto.randomUUID())
			if (pid == 0) {
				throw `ERROR: failed to run autorack on ${this.hostname}, availRAM = ${this.availRam}`
			}
			return new Task(this, pid, "autorack", Date.now())
		}

		execAutoPwn() {
			const pid = ns.exec("/payloads/autopwn.js", this.hostname, 1, crypto.randomUUID())
			if (pid == 0) {
				throw `ERROR: failed to run autopwn on ${this.hostname}, availRAM = ${this.availRam}`
			}
			return new Task(this, pid, "autorack", Date.now())
		}
	}

	class ServerPool {
		constructor() {
			this.pool = []
		}

		get size() {
			return this.pool.length
		}

		get maxRam() {
			return this.pool.reduce((x, y) => x + y.maxRam, 0)
		}

		get usedRam() {
			return this.pool.reduce((x, y) => x + y.usedRam, 0)
		}

		get availRam() {
			return this.pool.reduce((x, y) => x + y.availRam, 0)
		}

		getHackCapacity() {
			return this.pool.reduce((x, y) => x + y.getHackCapacity(), 0)
		}

		getGrowCapacity() {
			return this.pool.reduce((x, y) => x + y.getGrowCapacity(), 0)
		}

		getWeakenCapacity() {
			return this.pool.reduce((x, y) => x + y.getWeakenCapacity(), 0)
		}

		killAll() {
			this.pool.forEach(server => ns.killall(server.hostname))
		}

		execHack(target, numThreads, split=true) {
			let tasks = []

			if (numThreads == 0) {
				throw new PoolException(`attempted to schedule task with threads = 0 (operation = hack, target = ${target})`)
			}

			for (const server of this.pool) {
				const capacity = server.getHackCapacity()
				if (capacity == 0) {
					// this server is full, so we skip it
					continue
				}
				else if(!split && capacity < numThreads) {
					// we don't want to split into multiple threads and this server doesn't have enough capacity, so we skip it
					continue
				}
				const newThreads = numThreads > capacity ? capacity : numThreads
				const task = server.execHack(target, newThreads)
				if (task.pid != 0) {
					tasks.push(task)
					// reduce the number of remaining threads by the number of thread started by the new task
					numThreads -= newThreads
					if (numThreads <= 0) {
						// all threads scheduled, we can stop assigning them now
						break
					}
				}
			}

			if (tasks.length == 0) {
				throw new PoolException(`attempted to schedule task, put pool has no capacity (operation = hack, target = ${target}, threads = ${numThreads})`)
			}
			return tasks
		}

		execGrow(target, numThreads, split=true) {
			let tasks = []

			if (numThreads == 0) {
				throw new PoolException(`attempted to schedule task with threads = 0 (operation = grow, target = ${target})`)
			}

			for (const server of this.pool) {
				const capacity = server.getGrowCapacity()
				if (capacity == 0) {
					// this server is full, so we skip it
					continue
				}
				else if(!split && capacity < numThreads) {
					// we don't want to split into multiple threads and this server doesn't have enough capacity, so we skip it
					continue
				}
				const newThreads = numThreads > capacity ? capacity : numThreads
				const task = server.execGrow(target, newThreads)
				if (task.pid != 0) {
					tasks.push(task)
					// reduce the number of remaining threads by the number of thread started by the new task
					numThreads -= newThreads
					if (numThreads <= 0) {
						// all threads scheduled, we can stop assigning them now
						break
					}
				}
			}

			if (tasks.length == 0) {
				throw new PoolException(`attempted to schedule task, put pool has no capacity (task = grow, target = ${target}, threads = ${numThreads})`)
			}
			return tasks
		}

		execWeaken(target, numThreads, split=true) {
			let tasks = []

			if (numThreads == 0) {
				throw new PoolException(`attempted to schedule task with threads = 0 (tasks = weaken, target = ${target} `)
			}

			for (const server of this.pool) {
				const capacity = server.getWeakenCapacity()
				if (capacity == 0) {
					// this server is full, so we skip it
					continue
				}
				else if(!split && capacity < numThreads) {
					// we don't want to split into multiple threads and this server doesn't have enough capacity, so we skip it
					continue
				}
				const newThreads = numThreads > capacity ? capacity : numThreads
				const task = server.execWeaken(target, newThreads)
				if (task.pid != 0) {
					tasks.push(task)
					// reduce the number of remaining threads by the number of thread started by the new task
					numThreads -= newThreads
					if (numThreads <= 0) {
						// all threads scheduled, we can stop assigning them now
						break
					}
				}
			}

			if (tasks.length == 0) {
				throw new PoolException(`attempted to schedule task, put pool has no capacity (task = weaken, target = ${target}, threads = ${numThreads})`)
			}
			return tasks
		}

		execAutoRack() {
			// TODO: skip if fully upgraded
			let tasks = []
			for (const server of this.pool) {
				if (server.availRam >= autoRackRamUsage) {
					const task = server.execAutoRack()
					if (task.pid != 0) {
						tasks.push(task)
						break
					}
				}
			}
			return tasks
		}

		execAutoPwn() {
			// TODO: skip if fully upgraded
			let tasks = []
			for (const server of this.pool) {
				if (server.availRam >= autoRackPwnUsage) {
					const task = server.execAutoPwn()
					if (task.pid != 0) {
						tasks.push(task)
						break
					}
				}
			}
			return tasks
		}

		execOptimalHack(target, factor = 1) {
			const numThreads = Math.min(this.getHackCapacity(), target.computeOptimalHackThreadCount() * factor)
			ns.print(`INFO: hacking ${target} with ${numThreads} threads`)
			return this.execHack(target, numThreads)
		}

		execOptimalGrow(target, factor = 1) {
			const numThreads = Math.min(this.getGrowCapacity(), target.computeOptimalGrowThreadCount() * factor)
			ns.print(`INFO: growing ${target} with ${numThreads} threads`)
			return this.execGrow(target, numThreads)
		}

		execOptimalWeaken(target, factor = 1) {
			const numThreads = Math.min(this.getWeakenCapacity(), target.computeOptimalWeakenThreadCount() * factor)
			ns.print(`INFO: weakening ${target} with ${numThreads} threads`)
			return this.execWeaken(target, numThreads)
		}

		refresh() {
			// recursively list all available servers on network
			const getAvailableServers = (hostnameToAnalyze = null, availableServers = []) => {
				for (const parentHostname of ns.scan(hostnameToAnalyze)) {
					const parentServer = new Server(parentHostname)
					// check if we alreay have added this server to the list
					if (!availableServers.map((n) => n.toString()).includes(parentServer.toString())) {
						// add to the list
						availableServers.push(parentServer)
						// add parent servers to the list
						availableServers = getAvailableServers(parentServer.hostname, availableServers)
					}
				}
				return availableServers
			}

			const availableServers = getAvailableServers()
			ns.print(`INFO: found ${availableServers.length} servers in network`)

			let serversToAdd = []
			availableServers
				.filter(server => server.hasRootAccess && server.maxRam > 0) // is able to run tasks
				.filter(server => !this.pool.find(poolServer => poolServer.hostname == server.hostname)) // we don't already have this server in the pool
				.filter(server => !loopSettings.poolBlacklist.find(blacklistedHostname => blacklistedHostname == server.hostname)) // hostname is blacklisted
				.forEach(server => serversToAdd.push(server))
			this.pool = this.pool.concat(serversToAdd)
			ns.print(`INFO: added ${serversToAdd.length} servers to pool`)

			return serversToAdd
		}

		prepare(target) {
			let tasks = []
			if (target.availMoney != target.maxMoney) {
				// schedule as many grow() as able or necessary
				const growThreads = Math.min(this.getGrowCapacity(), Math.ceil(ns.growthAnalyze(target.hostname, target.maxMoney / target.availMoney)))
				tasks = tasks.concat(this.execGrow(target, growThreads, true))
				// try to offset the security growth with some parallel weaken() as able
				const weakenThreads = Math.min(this.getWeakenCapacity(), Math.ceil(((target.currSecurityLevel - target.minSecurityLevel) + (growThreads * 0.004)) / 0.04)) // 0.05 is more optimal, but t doesn't leave a lot of margin
				if (weakenThreads > 0) {
					// it's okay to skip weaken() if we have no capacity, we will do it next cycle
					tasks = tasks.concat(this.execWeaken(target, weakenThreads, true))
				}
			}
			return new Batch(tasks, target)
		}

		attack(target) {
			if (target.availMoney != target.maxMoney) {
				// if money is not at maximum, grow()
				return this.execOptimalGrow(target)
			}
			else if (target.currSecurityLevel != target.minSecurityLevel) {
				// if security is not at minimum, weaken()
				return this.execOptimalWeaken(target)
			}
			else {
				// if money is at maximum and security is at minimum, hack()
				return this.execOptimalHack(target)
			}
		}
	}

	class Task {
		constructor(server, pid, operation, completionTime, target = null) {
			this.server = server
			this.pid = pid
			this.operation = operation
			this.completionTime = completionTime
			this.target = target
		}

		isRunning() {
			return this.server.pids.includes(this.pid)
		}

		getTimeUntilComplete() {
			return Math.max(0, this.completionTime - Date.now())
		}

		async sleepUntilCompleted() {
			while (this.isRunning()) {
				const sleepTime = this.getTimeUntilComplete() + 200
				ns.print(`INFO: sleeping ${sleepTime} milliseconds for pid ${this.pid} on server ${this.server} to complete (op = ${this.operation}, target = ${this.target})`)
				await ns.sleep(sleepTime)
			}
		}
	}

	class Batch {
		constructor(tasks, target = null) {
			this.tasks = tasks
			this.target = target
		}

		get completionTime() {
			return Math.max(...this.tasks.map(t => t.completionTime))
		}

		isRunning() {
			return this.tasks.reduce((r, t) => r || t.isRunning(), false)
		}

		getTimeUntilComplete() {
			return Math.max(...this.tasks.map(t => t.getTimeUntilComplete()))
		}

		async sleepUntilCompleted() {
			while (this.isRunning()) {
				const sleepTime = this.getTimeUntilComplete() + 200
				ns.print(`INFO: sleeping ${sleepTime} milliseconds for batch to complete (tasks = [${this.tasks.map(t => t.operation).join(",")}], target = ${this.target})`)
				await ns.sleep(sleepTime)
			}
		}
	}

	// ns.exec("map.js", "home")

	const loop = async () => {
		const pool = new ServerPool()
		let tasks = []
		let preparingTargets = []
		let targets = []
		let lastRefreshLevel = ns.getHackingLevel()
		let refresh = true
		let reScore = false

		const expFarmTarget = new Server(loopSettings.expTargetHostname)
		expFarmTarget.targetOf = 0

		ns.atExit(() => pool.killAll())

		// run autopwn once on startup to enable new servers
		ns.exec("/payloads/autopwn.js", "home")
		await ns.sleep(500)

		while (true) {
			// every 10 level-up refresh the list of targets and recalculate the target priorities
			if (ns.getHackingLevel() - lastRefreshLevel >= 10) {
				ns.print(`INFO: triggering auto-refresh of targets`)
				refresh = true
				reScore = true
				lastRefreshLevel = ns.getHackingLevel()
			}

			// refresh targets
			if (refresh) {
				pool.refresh()
				tasks.concat(pool.execAutoPwn())
				pool.pool
					.filter(target => !preparingTargets.includes(target))
					.filter(target => !targets.includes(target))
					.filter(target => target.hackable) // server is hackable
					.filter(target => loopSettings.moneyTargetsWhitelist.length != 0 ? loopSettings.moneyTargetsWhitelist.includes(target.hostname) : true)
					.filter(target => !loopSettings.moneyTargetsBlacklist.includes(target.hostname))
					.forEach(target => {
						target.targetOf = 0
						ns.print(`INFO: added ${target} as potential target`)
						preparingTargets.push(target)
					})
				refresh = false
				await ns.sleep(500)
			}

			// prepare targets
			// TODO: this is currently very expensive because we need to place all target in ideal conditions before computing the score
			// in the future, calculate score and prepare only relevant targets
			let tmpPreparingTargets = [] // to avoid concurrent modification
			for (const target of preparingTargets) {
				try {
					if (target.targetOf == 0) {
						const prepareBatch = pool.prepare(target)
						if (prepareBatch.tasks.length == 0) {
							// target is ready to go
							// target.targetData.operation = "ready"
							targets.push(target)
							reScore = true // trigger a reScore
						}
						else {
							// track the prepare batch
							tasks.push(prepareBatch)
							target.targetOf += 1
							tmpPreparingTargets.push(target)
						}
					}
					else {
						// we can't schedule more tasks, but keep track of the target
						tmpPreparingTargets.push(target)
					}

				}
				catch (e) {
					if (e instanceof PoolException) {
						ns.print(e.toString())
						tmpPreparingTargets.push(target) // TODO: this is very jank
						if (e.fatal) {
							throw e
						}
					}
					else {
						throw e
					}
				}
			}
			preparingTargets = tmpPreparingTargets

			// score targets
			if (reScore) {
				ns.print(`INFO: sorting ${targets.length} targets`)
				const score = (target) => target.maxMoney * target.hackChance / target.growTime
				targets.sort((t1, t2) => score(t2) - score(t1))
				reScore = false
			}

			// extract money
			if (targets.length != 0) {
				for (let i = 0; i < Math.min(maxTargets, targets.length); i++) {
					try {
						const target = targets[i]
						if (target.targetOf == 0) {
							// ns.print(`hacking ${target}`)
							// ns.print(`maxMoney: ${target.maxMoney}`)
							// ns.print(`availMoney: ${target.availMoney}`)
							// ns.print(`minSecurity: ${target.minSecurityLevel}`)
							// ns.print(`currSecurity: ${target.currSecurityLevel}`)
							const newTasks = pool.attack(target)
							tasks = tasks.concat(newTasks)
							target.targetOf += newTasks.length
						}
					}
					catch (e) {
						if (e instanceof PoolException) {
							ns.print(e.toString())
						}
						else {
							throw e
						}
					}
				}
			}

			// farm exp
			if (expFarmTarget.targetOf == 0 && loopSettings.expMaxAvailRamUsePercent != 0) {
				// we want to avoid filling up the ram to not starve other tasks
				const usableRam = pool.availRam * loopSettings.expMaxAvailRamUsePercent
				let newTasks = []
				try {
					if (expFarmTarget.currSecurityLevel > expFarmTarget.minSecurityLevel) {
						// minimize security to speed up hack()
						newTasks = pool.execWeaken(expFarmTarget, Math.floor(usableRam / weakenRamUsage))
					}
					else {
						// hack() nonstop
						newTasks = pool.execHack(expFarmTarget, Math.floor(usableRam / hackRamUsage))
					}
					expFarmTarget.targetOf += newTasks.length
					tasks = tasks.concat(newTasks)
				}
				catch (e) {
					// ns.print(e.toString())
					if (!e instanceof PoolException) {
						throw e
					}
				}
			}

			// sleep until new tick
			// ns.print(tasks)
			// ns.print(targets)
			// ns.print(preparingTargets)
			ns.print(`pool = ${pool.size}, RAM = ${ns.formatRam(pool.usedRam)}/${ns.formatRam(pool.maxRam)} (${ns.formatPercent(pool.usedRam / pool.maxRam)}), targets = ${targets.length}/${targets.length + preparingTargets.length}, tasks = ${tasks.length}, refreshCountDown = ${10 - (ns.getHackingLevel() - lastRefreshLevel)}`)
			if (tasks.length > 0) {
				tasks.sort((t1, t2) => t1.completionTime - t2.completionTime)
				await tasks[0].sleepUntilCompleted()
				while (tasks.length != 0 && !tasks[0].isRunning()) {
					const task = tasks.shift()
					task.target.targetOf--
				}
			}
			else {
				throw "idle"
			}
		}
	}
	await loop()
}