/** @param {NS} ns */
export async function main(ns) {

	ns.clearLog()
	ns.tail()
	ns.disableLog('ALL')

	const getServers = (hostnameToAnalyze = "home", servers = ["home"]) => {
		for (const hostname of ns.scan(hostnameToAnalyze)) {
			// check if we alreay have added this server to the list
			if (!servers.includes(hostname)) {
				// add to the list
				servers.push(hostname)
				// map the host
				servers = getServers(hostname, servers)
			}
		}
		return servers
	}

	const solvers = new Map([
		["Find Largest Prime Factor", (data) => {
			const factorize = (n) => {
				let results = [];
				for (let i = 2; i <= n; i++) {
					// check if divisible
					if (n % i == 0) {
						// divide the number to test
						n = n / i;
						// store the divisor
						results.push(i);
					}
				}
				return results;
			}
			return factorize(data).pop()
		}],
		["Subarray with Maximum Sum", (data) => {
			let result = 0
			// loop 1: walk starting index
			for (let i = 0; i < data.length; i++) {
				let sum = 0
				// loop 2: sum everything from the stating index
				for (let j = i; j < data.length; j++) {
					sum += data[j]
					// if at any point during computing the sum we find a greater sum
					if (sum > result) {
						result = sum
					}
				}
			}
			return result
		}],
		["Total Ways to Sum", (data) => {
			// This can be solved with using the Partition function P
			// solution = P(n) - 1
			// https://en.wikipedia.org/wiki/Partition_function_(number_theory)
			// https://rosettacode.org/wiki/Partition_function_P
			const p = n => {
				var a = new Array(n + 1)
				a[0] = 1n

				for (let i = 1; i <= n; i++) {
					a[i] = 0n
					for (let k = 1, s = 1; s <= i;) {
						a[i] += (k & 1 ? a[i - s] : -a[i - s])
						k > 0 ? (s += k, k = -k) : (k = -k + 1, s = k * (3 * k - 1) / 2)
					}
				}

				return a[n]
			}
			return Number(p(data) - 1n)
		}],
		["Total Ways to Sum II", (data) => {
			// https://www.geeksforgeeks.org/coin-change-dp-7/
			// https://www.geeksforgeeks.org/understanding-the-coin-change-problem-with-dynamic-programming/
			const target = data[0]
			const numberPool = data[1]
			let memoizationTable = new Array(target + 1).fill(0)

			// there is always 1 way of returning 0 sum; with 0 number
			memoizationTable[0] = 1

			// loop through available number in the pool
			for (let i = 0; i < numberPool.length; i++) {
				// compute each entry in the memoization table
				for (let j = 0; j < memoizationTable.length; j++) {
					// ensure we do not lookup negative indices
					if (numberPool[i] <= j) {
						// compute the value at this index and store in the table to be referenced next iteration
						memoizationTable[j] += memoizationTable[j - numberPool[i]]
					}
				}
			}
			// last entry in the table is our answer
			return memoizationTable.pop()
		}],
		["Spiralize Matrix", (data) => {
			const matrix = data

			let resultSize = matrix.length * matrix[0].length
			let result = []

			let topStop = 1
			let bottomStop = matrix.length - 1
			let leftStop = 0
			let rightStop = matrix[0].length - 1

			let pos = [0, 0]
			let vector = [1, 0]

			while (result.length < resultSize) {
				if (vector[0] == 1 && vector[1] == 0) {
					// going right
					while (pos[1] <= rightStop) {
						result.push(matrix[pos[0]][pos[1]])
						pos[1]++
					}
					rightStop--
					pos[0]++
					pos[1]--
					// go down next
					vector = [0, 1]
				}
				else if (vector[0] == -1 && vector[1] == 0) {
					// going left
					while (pos[1] >= leftStop) {
						result.push(matrix[pos[0]][pos[1]])
						pos[1]--
					}
					leftStop++;
					// go up next
					pos[0]--
					pos[1]++
					vector = [0, -1]
				}
				else if (vector[0] == 0 && vector[1] == 1) {
					// going down
					while (pos[0] <= bottomStop) {
						result.push(matrix[pos[0]][pos[1]])
						pos[0]++
					}
					bottomStop--
					// go right next
					pos[0]--
					pos[1]--
					vector = [-1, 0]
				}
				else if (vector[0] == 0 && vector[1] == -1) {
					// going up
					while (pos[0] >= topStop) {
						result.push(matrix[pos[0]][pos[1]])
						pos[0]--
					}
					topStop++
					// go left next
					pos[0]++
					pos[1]++
					vector = [1, 0]
				}
				else {
					ns.print("Invalid state")
					break
				}
			}
			return result
		}],
		["Array Jumping Game", (data) => {
			return solvers.get("Array Jumping Game II")(data) == 0 ? 0 : 1
		}],
		["Array Jumping Game II", (data) => {
			let indexJumpSums = [] // contain an array of to index this index can jump
			let position = 0
			let moves = [position]
			let goal = data.length - 1
			for (let i = 0; i < data.length; i++) {
				indexJumpSums[i] = data[i] + i
			}

			do {
				position = moves[moves.length - 1]
				const jumpValue = data[position]

				if (goal - position <= jumpValue) { // check if the goal is within reach
					moves.push(goal) // reached the goal
					break
				}
				else {
					// find the largest jump possible at our current position
					let maxJump = position
					for (let i = position; i <= position + jumpValue; i++) {
						if (indexJumpSums[i] > indexJumpSums[maxJump]) {
							maxJump = i // this index allows us to jump the furthest
						}
					}
					if (maxJump == position) { // we haven't moved, we are stuck at this index
						moves = [0] // impossible!
						break
					}
					moves.push(maxJump)
				}
			} while (position != goal)

			return moves.length - 1
		}],
		["Merge Overlapping Intervals", (data) => {
			let result = []
			let newInterval = []
			// for this algo to work, we need to sort the input by the min of the intervals
			data.sort((a, b) => a[0] - b[0])
			for (let i = 0; i < data.length; i++) {
				// first interval
				if (newInterval.length == 0) {
					newInterval = data[i]
				}
				// if the min of the current inverval is less than the max of the newInterval
				if (data[i][0] <= newInterval[1]) {
					newInterval[1] = Math.max(data[i][1], newInterval[1]) // set the largest of the max to be the max of the newInterval
				}
				else {
					result.push(newInterval) // we are out of the new interval so push it to the result
					newInterval = data[i] // start a new newInterval
				}
				// last interval
				if (i == data.length - 1) {
					newInterval[1] = Math.max(data[i][1], newInterval[1])
					result.push(newInterval)
				}
			}
			return result
		}],
		["Generate IP Addresses", (data) => {
			Array.prototype.swap = function (a, b) {
				this[a] = this.splice(b, 1, this[a])[0];
				return this;
			}

			// https://en.wikipedia.org/wiki/Permutation#Generation_in_lexicographic_order
			const permute = (a) => {
				// 1. Find the largest index k such that a[k] < a[k + 1]. If no such index exists, the permutation is the last permutation.
				let k = a.length - 2
				while (a[k] >= a[k + 1]) {
					k--
					if (k < 0) {
						return false
					}
				}
				// 2. Find the largest index l greater than k such that a[k] < a[l].
				let l = a.length - 1
				while (a[k] >= a[l]) {
					l--
				}
				// 3. Swap the value of a[k] with that of a[l].
				a.swap(k, l);
				// 4. Reverse the sequence from a[k + 1] up to and including the final element a[n].
				const length = a.length - (k + 1)
				for (let i = 0; i < length / 2; i++) {
					a.swap(k + 1 + i, a.length - i - 1)
				}
				return true
			}

			const genIp = (data, digitPermutation) => {
				let octets = []
				let i = 0
				for (const digitCount of digitPermutation) {
					let newOctet = data.slice(i, i + digitCount)
					// sanity tests
					if (parseInt(newOctet) > 255) {
						return null
					}
					if (newOctet.length != 1 && newOctet[0] === "0") {
						return null
					}
					octets.push(newOctet)
					i += digitCount
				}
				return octets.join(".")
			}

			// build the couple first permutations
			// dumb, but it works
			let basePermutations = []
			switch (data.length) {
				case 4:
					basePermutations = [
						[1, 1, 1, 1],
					]
					break
				case 5:
					basePermutations = [
						[1, 1, 1, 2],
					]
					break
				case 6:
					basePermutations = [
						[1, 1, 1, 3],
						[1, 1, 2, 2],
					]
					break
				case 7:
					basePermutations = [
						[1, 1, 2, 3],
						[1, 2, 2, 2],
					]
					break
				case 8:
					basePermutations = [
						[1, 2, 2, 3],
						[1, 1, 3, 3],
						[2, 2, 2, 2],
					]
					break
				case 9:
					basePermutations = [
						[2, 2, 2, 3],
						[1, 2, 3, 3],
					]
					break
				case 10:
					basePermutations = [
						[2, 2, 3, 3],
						[1, 3, 3, 3],
					]
					break
				case 11:
					basePermutations = [
						[2, 3, 3, 3],
					]
					break
				case 12:
					basePermutations = [
						[3, 3, 3, 3],
					]
					break
			}

			// permute
			let digitPermutations = []
			for (let i = basePermutations.length - 1; i >= 0; i--) {
				let currentDigitPermutation = basePermutations[i]
				do {
					// keep a copy of the current permutation
					digitPermutations.push(currentDigitPermutation.slice(0))
				} while (permute(currentDigitPermutation)) // permute until we are done
			}

			// generate ips
			let ips = []
			for (const digitPermutation of digitPermutations) {
				const ip = genIp(data, digitPermutation)
				if (ip != null) {
					ips.push(ip)
				}
			}

			return ips
		}],
		["HammingCodes: Integer to Encoded Binary", (data) => {
			const dataBits = data.toString(2).split("").map(x => parseInt(x))

			// compute the number of parity bits
			let parityBitsCount = 0
			while ((1 << parityBitsCount) < (dataBits.length + parityBitsCount + 1)) {
				parityBitsCount++
			}

			// initialize encodedData and setup the data at the correct indices
			let encodedData = new Array(dataBits.length + parityBitsCount)
			let dataIndex = 0
			for (let i = 0; i < encodedData.length; i++) {
				// TODO: why this works?
				if (Math.log2(i + 1) % 1 == 0) { // check if i is power of 2
					// initialize parity bit
					encodedData[i] = 0
				}
				else {
					// set data bit
					encodedData[i] = dataBits[dataIndex++]
				}
			}

			// set the parity bits
			for (let i = 0; i < parityBitsCount; i++) { // loop every parity bits
				const parityIndex = (1 << i) - 1;
				let parity = 0;

				// xor together every bits this parity bit covers
				for (let j = parityIndex; j < encodedData.length; j += (parityIndex + 1) * 2) {
					for (let k = j; k < j + parityIndex + 1; k++) {
						parity ^= encodedData[k];
					}
				}

				// set parity bit
				encodedData[parityIndex] = parity;
			}

			// insert the overall parity bit at index 0
			encodedData.unshift(encodedData.reduce((x, y) => x ^ y))

			// reduce encodedData to a string of 1s and 0s
			return encodedData.reduce((x, y) => x + y, "")
		}],
		["HammingCodes: Encoded Binary to Integer", (data) => {
			// https://www.youtube.com/watch?v=b3NxrZOu_CE

			let encodedData = data.split("").map(x => parseInt(x))

			// NOTE: bit at index 0 is the overall parity bit
			// we will not get a string with 2 errors, so can ignore it

			// check for flipped bit and correct it
			let errorBitIndex = 0
			for (let i = 0; i < encodedData.length; i++) {
				if (encodedData[i] == 1) {
					errorBitIndex ^= i
				}
			}
			if (errorBitIndex != 0) {
				encodedData[errorBitIndex] ^= 1
			}

			// extract data
			let dataBits = []
			for (let i = 0; i < encodedData.length; i++) {
				if ((i & (i - 1)) != 0 && i != 0) { // check if i is not power of 2
					// extract data bit
					dataBits.push(encodedData[i])
				}
			}

			return parseInt(dataBits.reduce((x, y) => x + y, ""), 2).toString(10)
		}],
		["Encryption I: Caesar Cipher", (data) => {
			const charList = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z']
			const plaintext = data[0].split("")
			const key = data[1]

			return plaintext.map(c => {
				const i = charList.indexOf(c)
				if (i >= 0) {
					return charList[(i + (26 - key)) % 26]
				}
				else {
					return ' '
				}
			}).reduce((x, y) => x + y, "")

		}],
		["Encryption II: Vigenère Cipher", (data) => {
			const charList = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z']
			const cypher = (a, b) => charList[(charList.indexOf(a) + charList.indexOf(b)) % 26]
			const plaintext = data[0].split("")
			const key = data[1].split("")

			let answer = ""
			let keyIndex = 0
			for (const char of plaintext) {
				answer += cypher(char, key[keyIndex++ % key.length])
			}
			return answer
		}]
	])

	// ns.codingcontract.createDummyContract("Find Largest Prime Factor") // ok
	// ns.codingcontract.createDummyContract("Subarray with Maximum Sum") // ok
	// ns.codingcontract.createDummyContract("Total Ways to Sum") // ok
	// ns.codingcontract.createDummyContract("Total Ways to Sum II") // ok
	// ns.codingcontract.createDummyContract("Spiralize Matrix") // ok
	// ns.codingcontract.createDummyContract("Array Jumping Game") // ok
	// ns.codingcontract.createDummyContract("Array Jumping Game II") //ok
	// ns.codingcontract.createDummyContract("Merge Overlapping Intervals") // ok
	// ns.codingcontract.createDummyContract("Generate IP Addresses") // ok
	// ns.codingcontract.createDummyContract("Algorithmic Stock Trader I")
	// ns.codingcontract.createDummyContract("Algorithmic Stock Trader II")
	// ns.codingcontract.createDummyContract("Algorithmic Stock Trader III")
	// ns.codingcontract.createDummyContract("Algorithmic Stock Trader IV")
	// ns.codingcontract.createDummyContract("Minimum Path Sum in a Triangle")
	// ns.codingcontract.createDummyContract("Unique Paths in a Grid I")
	// ns.codingcontract.createDummyContract("Unique Paths in a Grid II")
	// ns.codingcontract.createDummyContract("Shortest Path in a Grid")
	// ns.codingcontract.createDummyContract("Sanitize Parentheses in Expression")
	// ns.codingcontract.createDummyContract("Find All Valid Math Expressions")
	// ns.codingcontract.createDummyContract("HammingCodes: Integer to Encoded Binary") // ok
	// ns.codingcontract.createDummyContract("HammingCodes: Encoded Binary to Integer") // ok
	// ns.codingcontract.createDummyContract("Proper 2-Coloring of a Graph")
	// ns.codingcontract.createDummyContract("Compression I: RLE Compression")
	// ns.codingcontract.createDummyContract("Compression II: LZ Decompression")
	// ns.codingcontract.createDummyContract("Compression III: LZ Compression")
	// ns.codingcontract.createDummyContract("Encryption I: Caesar Cipher") // ok
	// ns.codingcontract.createDummyContract("Encryption II: Vigenère Cipher") // ok

	const server = "home"
	for (const contract of ns.ls(server, '.cct')) {
		const contractType = ns.codingcontract.getContractType(contract, server)
		ns.print(`found contract ${contract} of type "${contractType}" on server ${server}`)
		if (solvers.has(contractType)) {
			ns.print("attempting to solve")
			const answer = solvers.get(contractType)(ns.codingcontract.getData(contract, server))
			const reward = ns.codingcontract.attempt(answer, contract, server)
			if (reward == "") {
				ns.print(`ERROR: failed to solve contract! answer = ${answer}`)
				break
			}
			else {
				ns.print(reward)
			}
		}
		else {
			ns.print(`WARNING: unknown contract type: ${contractType}`)
		}
	}

	// for (const server of getServers()) {
	// 	for (const contract of ns.ls(server, '.cct')) {
	// 		const contractType = ns.codingcontract.getContractType(contract, server)
	// 		ns.print(`found contract ${contract} of type "${contractType}" on server ${server}`)
	// 		if (solvers.has(contractType)) {
	// 			const answer = solvers.get(contractType)(ns.codingcontract.getData(contract, server))
	// 			const reward = ns.codingcontract.attempt(answer, contract, server)
	// 			if (answer == "") {
	// 				ns.print(`ERROR: failed to solve contract! answer = ${answer}`)
	// 				break
	// 			}
	// 			else {
	// 				ns.print(reward)
	// 			}
	// 		}
	// 		else {
	// 			ns.print(`WARNING: unknown contract type: ${contractType}`)
	// 		}
	// 	}
	// }
}