/** @param {NS} ns */
export async function main(ns) {
	const moneyLowPercent = 0.75
	const securityHighOffset = 5

	const moneyFormat = new Intl.NumberFormat({ style: "currency" })

	const pwnMethods = []
	if (ns.fileExists("BruteSSH.exe")) {
		pwnMethods.push(ns.brutessh)
	}
	if (ns.fileExists("FTPCrack.exe")) {
		pwnMethods.push(ns.ftpcrack)
	}
	if (ns.fileExists("HTTPWorm.exe")) {
		pwnMethods.push(ns.httpworm)
	}
	if (ns.fileExists("relaySMTP.exe")) {
		pwnMethods.push(ns.relaysmtp)
	}
	if (ns.fileExists("SQLInject.exe")) {
		pwnMethods.push(ns.sqlinject)
	}

	// ns.clearLog()
	// ns.tail()
	// ns.disableLog("getServerMaxRam")
	// ns.disableLog("scp")
	// ns.disableLog("killall")
	ns.print(`${pwnMethods.length} pwning method(s) available!`)

	class Server {
		constructor(hostname) {
			this.hostname = hostname
			this.maxMoney = ns.getServerMaxMoney(hostname)
			this.minSecurityLevel = ns.getServerMinSecurityLevel(hostname)
			this.maxRam = ns.getServerMaxRam(hostname)
			this.requiredHackLevel = ns.getServerRequiredHackingLevel(hostname)
			this.hasRootAccess = ns.hasRootAccess(hostname)
			this.numPortRequired = ns.getServerNumPortsRequired(hostname)
		}

		toString() {
			return this.hostname
		}

		get usedRam() {
			return ns.getServerUsedRam(this.hostname)
		}

		get availRam() {
			return this.maxRam - this.usedRam
		}

		get availMoney() {
			return ns.getServerMoneyAvailable(this.hostname)
		}

		get currSecurityLevel() {
			return ns.getServerSecurityLevel(this.hostname)
		}

		get weakenTime() {
			return ns.getWeakenTime(this.hostname)
		}

		get hackTime() {
			return ns.getHackTime(this.hostname)
		}

		get growTime() {
			return ns.getGrowTime(this.hostname)
		}

		get score() {
			if (this.requiredHackLevel > ns.getHackingLevel() || this.maxMoney == 0) {
				return Infinity
			}
			else {
				const percentMinSecurityLevel = (this.minSecurityLevel / 100)
				const targetRequiredHackLevel = ns.getHackingLevel() * 0.33
				const percentLevelDeviation = Math.abs((this.requiredHackLevel - targetRequiredHackLevel) / targetRequiredHackLevel)
				// return Math.round(this.maxMoney * percentMinSecurityLevel * percentDeviation)
				return percentMinSecurityLevel + percentLevelDeviation
			}
		}

		computeScriptCapacity(script) {
			return Math.floor(this.maxRam / ns.getScriptRam(script))
		}

		pwn() {
			if (nodeNumPortsRequired <= pwnMethods.length) {
				ns.tprint(`pwning ${this.hostname}`)
				for (let i = 0; i < this.numPortRequired; i++) {
					pwnMethods[i](this.hostname)
				}
				ns.nuke(this.hostname)
				return true
			}
			else {
				ns.print(`missing requirements to pwn ${this.hostname}. (${pwnMethods.length} < ${this.numPortRequired})`)
				return false
			}
		}

		exploit(target) {
			const moneyLowThresh = target.maxMoney * moneyLowPercent
			const securityHighThresh = target.minSecurityLevel + securityHighOffset

			ns.print(`setup ${this}`)
			const capacity = this.computeScriptCapacity("exploit.js")
			if (capacity > 0) {
				ns.scp("exploit.js", this.hostname)
				ns.killall(this.hostname)
				ns.exec("exploit.js", this.hostname, capacity, target.hostname, moneyLowThresh, securityHighThresh)
			}
		}
	}

	/**
	 * Recursively list all available worker nodes
	 */
	const getAvailableNodes = (hostnameToAnalyze = "home", availableNodes = []) => {
		for (const parentHostname of ns.scan(hostnameToAnalyze)) {
			const parentServer = new Server(parentHostname)
			// check if we alreay have added this node to the list
			if (availableNodes.map((n) => n.toString()).includes(parentServer.toString())) {
				continue //skip this node
			}
			// if we are not root on the server yet
			if (!parentServer.hasRootAccess) {
				// gain access
				if (!parentServer.pwn()) {
					// failed, skip this node
				}
			}
			ns.print(`adding ${parentServer} to worker list`)
			availableNodes.push(parentServer)
			availableNodes = getAvailableNodes(parentServer.hostname, availableNodes)
		}
		return availableNodes
	}

	const printNodeTab = (nodes) => {
		const format = "%-15ss %10s %10s %10s %10s %10s %10s"
		ns.tprintf(format, "host", "score", "maxMoney", "minSecurity", "root?", "hackLevel", "maxRam")
		nodes.forEach((n) => {
			ns.tprintf(format, n.hostname, ns.formatNumber(n.score), ns.formatNumber(n.maxMoney), ns.formatNumber(n.minSecurityLevel), n.hasRootAccess, n.requiredHackLevel, ns.formatRam(n.maxRam))
		})
		ns.tprint(`total nodes: ${ns.formatNumber(nodes.length)}`)
		ns.tprint(`total RAM: ${ns.formatRam(nodes.reduce((x, y) => x + y.maxRam, 0))}`)
	}

	// find optimal target
	const availableNodes = getAvailableNodes()
	availableNodes.sort((a, b) => a.score - b.score)
	const target = availableNodes.at(0)

	// print status
	printNodeTab(availableNodes)
	ns.tprint(`targeting ${target}`)

	//execute
	availableNodes.forEach((n) => { n.exploit(target) })
}