import { getServerTree } from "../lib/ServerTree.js"

/** @param {NS} ns */
export async function main(ns) {
	const serverTree = getServerTree(ns)

	// 1st pass: recursively find servers to backdoor
	const getTreeOfServerToBackdoor = (subtree) => {
		const childrenToBackdoor = []
		for (const child of subtree["children"]) {
			const backdoorSubtreeOfChild = getTreeOfServerToBackdoor(child)
			if (backdoorSubtreeOfChild != null) {
				childrenToBackdoor.push(backdoorSubtreeOfChild)
			}
		}
		if (childrenToBackdoor.length != 0 || (!subtree.info.backdoorInstalled && subtree.info.hasAdminRights && subtree.info.requiredHackingSkill <= ns.getHackingLevel())) {
			return {
				"info": subtree.info,
				"children": childrenToBackdoor,
			}
		}
		else {
			return null
		}
	}

	// 2nd pass: walk the tree and backdoor servers
	const backdoor = async (subtree) => {
		if (!subtree.info.backdoorInstalled && subtree.info.requiredHackingSkill <= ns.getHackingLevel()) {
			ns.tprintf("backdoor %s, do not change server", subtree.info.hostname)
			await ns.singularity.installBackdoor()
			ns.tprint("backdoor complete")
		}
		for (const child of subtree.children) {
			// connect to child
			if (!ns.singularity.connect(child.info.hostname)) {
				ns.tprintf("failed to connect: %s", child.info.hostname)
				ns.exit()
			}
			// backdoor the child
			await backdoor(child)
			// connect back to current host
			if (!ns.singularity.connect(subtree.info.hostname)) {
				ns.tprintf("failed to connect: %s", subtree.info.hostname)
				ns.exit()
			}
		}
	}

	// execute
	const serverTreeToBackdoor = getTreeOfServerToBackdoor(serverTree)
	if (serverTreeToBackdoor == null) {
		ns.tprint("nothing to do")
	}
	else {
		await backdoor(serverTreeToBackdoor)
	}
	ns.tprint("script complete")
}