/** @param {NS} ns */
export async function main(ns) {
	const maxServerCount = ns.getPurchasedServerLimit()
	let serverCount = ns.getPurchasedServers().length

	ns.print(`we have ${serverCount} servers`)
	ns.print(`we can have ${maxServerCount} servers`)

	const getLowestServerRAM = (servers) => {
		return Math.min.apply(Math, servers.map(ns.getServerMaxRam))
	}

	if (serverCount < maxServerCount) {
		// create missing servers
		const requestedRAM = 2
		const serverCost = ns.getPurchasedServerCost(requestedRAM)
		ns.print(`server cost is ${ns.formatNumber(serverCost)}`)

		while (serverCount < maxServerCount) {
			const serverName = `worker-${serverCount}`
			if (ns.getServerMoneyAvailable("home") >= serverCost) {
				ns.tprint(`racking server ${serverName}`)
				ns.purchaseServer(serverName, requestedRAM)
				serverCount++
			}
			else {
				break
			}
		}
	}

	if (serverCount == maxServerCount) {
		const servers = ns.getPurchasedServers()
		let success = true
		do {
			// upgrade servers
			const requestedRAM = getLowestServerRAM(servers) * 2
			for (const serverName of servers) {
				const serverRAM = ns.getServerMaxRam(serverName)
				if (serverRAM < requestedRAM) {
					const serverCost = ns.getPurchasedServerUpgradeCost(serverName, requestedRAM)
					ns.print(`selected server ${serverName} to be upgraded (current RAM: ${ns.formatRam(serverRAM)}, upgraded RAM: ${ns.formatRam(requestedRAM)}, upgrade cost: ${ns.formatNumber(serverCost)})`)
					if (ns.getServerMoneyAvailable("home") >= serverCost) {
						ns.tprint(`upgrading server ${serverName}`)
						ns.killall(serverName) //TODO: make it more graceful
						ns.upgradePurchasedServer(serverName, requestedRAM)
					}
					else {
						success = false
					}
				}
			}
		} while (success)
	}
}